//******************************* 615driver.c ****************************
//* This file contains loadable Linux device driver for Sensoray Model 615
//*
//* Autor:		 David Stroupe and Charlie X. Liu
//*
//* Revision:
//*	    2002   0.0   David Stroupe   initial
//*	06/16/03   0.1   Charlie X. Liu  fixes for caption and warnings
//*	06/20/03   0.1.1 Charlie X. Liu  patch for i845 chipset bug fix
//*	08/10/04   0.2   Charlie X. Liu  adjustment and bug fixes for 
//*	                                 videopresent detection
//*
//* Copyright (C) 2002-2004	 Sensoray Co., Inc.
//************************************************************************
/*
This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA
**************************************************************************/


#ifndef __KERNEL__
#  define __KERNEL__
#endif
#ifndef MODULE
#  define MODULE
#endif


#define DEBUG 1
//#undef DEBUG
#ifdef DEBUG
#define DBG(x...) printk(x)
#else
#define DBG(x...)
#endif

#define DEBUGI 1
//#undef DEBUGI
#ifdef DEBUGI
#define DBGI(x...) printk(x)
#else
#define DBGI(x...)
#endif

#define DEBUGM 1
#undef DEBUGM
#ifdef DEBUGM
#define DBGM(x...) printk(x)
#else
#define DBGM(x...)
#endif

//#define OOPSTRACE 1
//#ifdef OOPSTRACE
//#define EnterFunction(x)   printk("Enter: %s, %s line %i\n",x,__FILE__,__LINE__)
//#define LeaveFunction(x)   printk("Leave: %s, %s line %i\n",x,__FILE__,__LINE__)
//#else
//#define EnterFunction(x)   do {} while (0)
//#define LeaveFunction(x)   do {} while (0)
//#endif


#define WR7146(index, offset, data) writel(data, board[index].base + offset)
#define RR7146(index, offset) readl( board[index].base + offset)


#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/wrapper.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/proc_fs.h>
#include <linux/pci.h>
#include <linux/version.h>
#include <linux/delay.h>
#include <linux/fs.h>
#include <linux/ioctl.h>
#include <linux/time.h>
#include <asm/io.h>
#include <asm/uaccess.h>
#include "615.h"
#include "615driver.h"

MODULE_DESCRIPTION("615driver - Sensoray Model 615 loadable Linux device driver module");
MODULE_AUTHOR("David Stroupe");
MODULE_LICENSE("GPL");

static int iocontrol (struct inode *inode, struct file *file, unsigned int cmd, unsigned long arg);
static int S615_Release(struct inode* inode, struct file* file);
static int S615_Open(struct inode* inode, struct file* file);
static int S615_Mmap (struct file *flip, struct vm_area_struct *vma);
static ssize_t S615_Read (struct file *flip, char* buffer, size_t size, loff_t offset);
static ssize_t S615_Write (struct file *flip, const char* buffer, size_t size, loff_t offset);
static BYTE ReadGA (int index, WORD addr);
static void WR060 (int index, WORD addr, BYTE bdata);
static void WriteGA (int index, WORD addr, BYTE data);
static ECODE InitI2C (DWORD index);
static void Init7146 (int index);
static ECODE InitXP (int index);
static void WriteSRAM (int index, WORD addr, BYTE data);
static BYTE ReadSRAM (int index, WORD addr);
static ECODE LoadSRAM (int index);
static void WaitDEBIDone (int index);
static void Reset060 (int index);
static ECODE Init060 (int index, MODE * pmode);
static ECODE Load060 (int index, MODE * pmode);
static ECODE LoadXP (int index, DWORD xpdata);
static ECODE WriteI2C (int index, BYTE devaddr, BYTE addr, BYTE bdata);
static ECODE ReadI2C (int index, BYTE devaddr, BYTE addr, BYTE * pdata);
static void SetNibble (BYTE nibblenumber, BYTE data, WORD * pWord);
static ECODE SetXP (int index, BYTE out, BYTE inp);
static ECODE SetMode (cardinfo* boardinfo);
static void CreateRPS (bufinfo* bf, MODE* pmode);
static void StartRPS (int index,bufinfo* bf, int jpegon);
static BYTE RR060 (int index, WORD addr);
static int GetXP (cardinfo* boardinfo, int chan);
static void CreateDIBHeader (bufinfo* bf, MODE * pMode);
static void GetChanInp(cardinfo* boardinfo , int* chan,  int* inp);
static int GetAvailableBuffer(cardinfo* boardinfo);
static void WriteCaption (cardinfo* boardinfo, int chan, char *pCaption);


static struct pci_device_id S615_id_tbl[] __devinitdata = {
  {0x1131, 0x7146, 0x6000, 0x0615, 0, 0, 0},
  {0,}
};

static struct pci_device_id i845DB_pci_tbl[] = {
  { 0x8086, 0x244e, 0, 0, 0, 0, 0 },
  {0,}
};				//for Intel i845 82801BA/82801DB

static struct pci_device_id i845BA_pci_tbl[] = {
  { 0x8086, 0x2448, 0, 0, 0, 0, 0 },
  {0,}
};				//for Intel i845 82801BA (some)

static struct pci_dev *dev_i845 = NULL;
static u32            i845_ConfigSpace_0x50;


MODULE_DEVICE_TABLE (pci, S615_id_tbl);

//#define P_INTEN   0x00DC
//#define P_INTSTAT 0x010C
//#define P_PSR     0x0110
#define IRQ_MASK 0x00000020


static struct file_operations S615_ops = {
  owner:THIS_MODULE,
  open:  S615_Open,
  read:  S615_Read,
  write: S615_Write,
  release:  S615_Release,
  mmap:  S615_Mmap,
  ioctl: iocontrol,
};

static char driverdevice[] = "s615";

static int S615_cards_found = 0;
static int devicemajornumber = 0;

static cardinfo board[SYS_BOARDS];
static int S615_RequestIRQ (cardinfo * board);
static int InitCard(int);

static int addpagetolist(struct pagelist* plist, unsigned long paddress)
{
  int numofpages = 0;
  struct pagelist *last, *new;
  last = plist;//start at the beginning of the list
  while(last->pnextpage != 0)
    {
      last = last->pnextpage;
      numofpages ++;
    }
  if(!last->physicaladdress)
    {
      last->physicaladdress = paddress;
      numofpages++;
    }
  else
    {
      numofpages++;//count the first page
      new = (struct pagelist*)kmalloc(sizeof(struct pagelist), GFP_KERNEL);
      memset(new, 0, sizeof(struct pagelist));
      last->pnextpage = new;
      new->physicaladdress = paddress;
      numofpages++;//count the new page
    }
  DBGM("Added address %lx to list, pages stored = %d\n", paddress, numofpages);
  return numofpages;
}

static int removepagefromlist(struct pagelist* plist, unsigned long *paddress)
{
  int numofpages = 0;
  struct pagelist *last, *prior;
  if(plist->physicaladdress == 0)//if the first page is empty
    return -1;
  prior = last = plist;//start from the beginning
  if(!prior->pnextpage)
    {
      *paddress = prior->physicaladdress;
      prior->physicaladdress = 0;
      DBGM("Returning address %lx %d pages remain stored in list\n", *paddress, numofpages);
      return numofpages;
    }
else
  numofpages++;//count the first page
  while(last->pnextpage !=0)
    {
      prior = last;
      last = last->pnextpage;
      numofpages++;
    }
  *paddress = last->physicaladdress;
  kfree(last);
  numofpages--;
  prior->pnextpage = 0;
  DBGM("Returning address %lx %d pages remain stored in list\n", *paddress, numofpages);
  return numofpages;
}

//used for debugging linked list of pages
static void walklist(struct pagelist* plist)
{
  struct pagelist *prior, *next;
  prior = plist;
  DBGM("walking list plist = %p\n", plist);
  do
    {
      next = prior->pnextpage;
      DBGM("This page's address = %p, the pa on this page = %lx, the next page =%p\n", prior, prior->physicaladdress, prior->pnextpage);
      if(next)
	{
	  DBGM("The page's address = %p, the pa on that page = %lx, next page after that =%p\n", next, next->physicaladdress, next->pnextpage);
	  prior = next;
	}
    }
  while(next);
}
static void GetChanInp(cardinfo* boardinfo , int* chan,  int* inp)
{
  int *pi = &boardinfo->sequenceposition;
  //sequenceposition is pointing to the correct place in the sequence
  *chan = (int)boardinfo->cardstatus.SEQUENCE[*pi][0];
  *inp = (int)boardinfo->cardstatus.SEQUENCE[*pi][1];
  printk("GetChanInp position = %i chan = %i, camera = %i\n", *pi, *chan, *inp);
  if(*pi == SEQ_MAX || boardinfo->cardstatus.SEQUENCE[++*pi][1] == 0 )//camera #0 is a terminator
    *pi=0;


}
static void DisableInterrupts(cardinfo* info)
{
  writel(0, info->base + P_INTEN);
  writel(0xffffffff, info->base + P_INTSTAT);
}

static void EnableInterrupts(cardinfo* info)
{
      writel(INTSTATUS_VSYNC, info->base + P_INTEN);
}

static ssize_t S615_Write (struct file *flip, const char* buffer, size_t size, loff_t offset)
{
  int result = 0;
  int i;
  cardinfo* boardinfo = ((cardinfo*)flip->private_data);
  userstatus *pstatus = kmalloc(sizeof(userstatus), GFP_KERNEL);
  if(!pstatus)
    return -ENOMEM;
  result = copy_from_user(pstatus, (userstatus*)buffer, sizeof(userstatus));
  if(result)
    {
      kfree(pstatus);
      return -EFAULT;
    }
  boardinfo->cardstatus.jpegon = pstatus->jpegon;
  boardinfo->cardstatus.captionmode = pstatus->captionmode;
  boardinfo->cardstatus.interlacemode = pstatus->interlacemode;
  for (i = 0; i < SEQ_MAX; i++)
    {
      boardinfo->cardstatus.SEQUENCE[i][0] = pstatus->SEQUENCE[i][0];
      boardinfo->cardstatus.SEQUENCE[i][1] = pstatus->SEQUENCE[i][1];
    }
  boardinfo->mode.format = pstatus->mode.format;	
  boardinfo->mode.bmpflip = pstatus->mode.bmpflip;	
  boardinfo->mode.bmpcolor = pstatus->mode.bmpcolor;	
  boardinfo->mode.bmpsize = pstatus->mode.bmpsize;	
  boardinfo->mode.jpgsize = pstatus->mode.jpgsize;	
  boardinfo->mode.captionformat = pstatus->mode.captionformat; 	
  boardinfo->mode.captionpos = pstatus->mode.captionpos; 	
  boardinfo->mode.fontsize = pstatus->mode.fontsize; 	
  boardinfo->mode.compfactor = pstatus->mode.compfactor; 	
  boardinfo->mode.xpinit = pstatus->mode.xpinit; 


  printk("boardinfo->mode.format = %lu\n", boardinfo->mode.format);	
  printk("boardinfo->mode.bmpflip = %lu\n", boardinfo->mode.bmpflip);	
  printk("boardinfo->mode.bmpcolor = %lu\n", boardinfo->mode.bmpcolor);	
  printk("boardinfo->mode.bmpsize = %lu\n", boardinfo->mode.bmpsize);	
  printk("boardinfo->mode.jpgsize = %lu\n", boardinfo->mode.jpgsize);	
  printk("boardinfo->mode.captionformat = %lu\n", boardinfo->mode.captionformat);	
  printk("boardinfo->mode.captionpos = %lu\n", boardinfo->mode.captionpos);	
  printk("boardinfo->mode.compfactor = %lu\n", boardinfo->mode.compfactor);	
  printk("boardinfo->mode.xpinit = %lu\n", boardinfo->mode.xpinit);	

  kfree(pstatus);
  return size;

}

static ssize_t S615_Read (struct file *flip, char* buffer, size_t size, loff_t offset)
{
  int bufnum = SYS_NOFREEBUFFERS;
  int result = 0;
  imgstatus *pstatus = kmalloc(sizeof(imgstatus), GFP_KERNEL);
  cardinfo* boardinfo =((cardinfo*)flip->private_data); 
  bufnum = GetAvailableBuffer(boardinfo);
  if(bufnum != SYS_NOFREEBUFFERS)
    {
      printk("Got a valid buffer %i\n", bufnum);
      pstatus->pjpg = (void*)boardinfo->bufferinfo[bufnum].jpegbufferuser;
      pstatus->pbmp = (void*)boardinfo->bufferinfo[bufnum].bmpbufferuser;
      pstatus->jpegsize = boardinfo->bufferinfo[bufnum].jpegsize;
      pstatus->camera = boardinfo->bufferinfo[bufnum].xpchan;
      pstatus->videopresent = boardinfo->bufferinfo[bufnum].videopresent;
      pstatus->buffernumber = bufnum;
      result = copy_to_user((imgstatus*)buffer, pstatus, sizeof(imgstatus));
      kfree(pstatus);
      if(!result)
	return sizeof(imgstatus);
      else
	return -EFAULT;
    }
  else  // no buffers available to read
    {
      kfree(pstatus);
      return 0;
    }
}

static int S615_Open(struct inode* inode, struct file* flip)
{
  //get the minor number and index through the cards to find the right one
  //the minor number will indicate which card in the system is being accessed
  //minor number = 0 will be the first card found in the system
  int minor;
  int error;
  int i;
  minor = MINOR(inode->i_rdev);
  if(board[minor].busy)
    return -EBUSY;
  else
    board[minor].busy = TRUE;
  if(!board[minor].pdev)
    {
      board[minor].busy = FALSE;
      return -ENODEV;
    }
  error = InitCard(minor);
  if(error)
    {
      board[minor].busy = FALSE;
      return error;
    }
  board[minor].ppagelist = (struct pagelist*)kmalloc(sizeof(struct pagelist), GFP_KERNEL);
  memset(board[minor].ppagelist, 0, sizeof(struct pagelist));
  board[minor].myminor = minor;
  flip->private_data =(void*)&board[minor];
  for(i = 0; i < SYS_BUFFERS; i++)
    {
      memset(&board[minor].bufferinfo[i], 0, sizeof(bufinfo));
   }

 


 board[minor].sequenceposition = 0;

  board[minor].cardstatus.SEQUENCE[0][0] = 0;
  board[minor].cardstatus.SEQUENCE[0][1] = 1;
  board[minor].cardstatus.SEQUENCE[1][0] = 1;
  board[minor].cardstatus.SEQUENCE[1][1] = 2;
  board[minor].cardstatus.SEQUENCE[2][0] = 2;
  board[minor].cardstatus.SEQUENCE[2][1] = 0;//terminator



  //everything is ok so
  MOD_INC_USE_COUNT;
  return 0;
}

static int S615_Release(struct inode* inode, struct file* flip)
{
  int i;
  cardinfo* boardinfo;
  unsigned int io_addr;
  unsigned int io_length;
  unsigned long paddress, pageaddy;
  DBG("S615_Release\n");
  boardinfo = (cardinfo*)flip->private_data;
  //disable interrupt
  DisableInterrupts(boardinfo);
  if(boardinfo->dointerrupts)
    free_irq (boardinfo->pdev->irq, boardinfo);
  boardinfo->dointerrupts = FALSE;
  boardinfo->wait_condition = KILLINTERRUPT;//now release any blocked I/O
  wake_up_interruptible (&(boardinfo->wait_queue));
  boardinfo->interruptcounter = 0;
  boardinfo->interruptslost = 0;
  boardinfo->wait_condition = 0;
  for (i = 0; i < SYS_BUFFERS ; i++)
    {
      if(boardinfo->bufferinfo[i].lpbmi)
	kfree(boardinfo->bufferinfo[i].lpbmi);
      if(boardinfo->bufferinfo[i].ptbmpuser)
	pci_free_consistent(boardinfo->pdev, 4000, (void*)boardinfo->bufferinfo[i].ptbmpuser, 
			    (dma_addr_t)boardinfo->bufferinfo[i].ptbmpphys);
      if(boardinfo->bufferinfo[i].ptjpguser)
	pci_free_consistent(boardinfo->pdev, 4000, (void*)boardinfo->bufferinfo[i].ptjpguser,
			    (dma_addr_t)boardinfo->bufferinfo[i]. ptjpgphys);
      if(boardinfo->bufferinfo[i].rpsbmpuser)
	pci_free_consistent(boardinfo->pdev, 4000, (void*)boardinfo->bufferinfo[i].rpsbmpuser, 
			    (dma_addr_t)boardinfo->bufferinfo[i].rpsbmpphys);
      if(boardinfo->bufferinfo[i].rpsjpguser)
	pci_free_consistent(boardinfo->pdev, 4000,(void*)boardinfo->bufferinfo[i]. rpsjpguser, 
			    (dma_addr_t)boardinfo->bufferinfo[i].rpsjpgphys);

    }
  if(boardinfo->ppagelist)
    {
      while(removepagefromlist(boardinfo->ppagelist, &paddress) != -1)
	{
	  pageaddy = (unsigned long)phys_to_virt(paddress);
	  mem_map_unreserve(virt_to_page(pageaddy));
	  free_page(pageaddy);
	}
      kfree(boardinfo->ppagelist);
    }
  iounmap ((void *) boardinfo->base);
  io_addr = pci_resource_start (boardinfo->pdev, 0);
  io_length = pci_resource_len (boardinfo->pdev, 0);
  release_mem_region (io_addr, io_length);
  MOD_DEC_USE_COUNT;
  boardinfo->busy = FALSE;

  return 0;
}

static int S615_Mmap (struct file *flip, struct vm_area_struct *vma)
{
  unsigned long             start, size, pageaddy, paddress;
 dma_addr_t                     ptbmpphys;	//bitmap page table start phys address;
 dma_addr_t                      rpsbmpphys;	//bitmap RPS prog start phys address;
  dma_addr_t                     ptjpgphys;	//jpeg page table start phys address;
  dma_addr_t                     rpsjpgphys;	//jpeg RPS prog start phys address;
  DWORD                     rpsbmpuser;	//bitmap RPS prog start user address;
  DWORD                     rpsjpguser;	//jpeg RPS prog start user address;
  DWORD                     ptbmpuser;	//bitmap page table start user address;
  DWORD                     ptjpguser;	//jpeg page table start user address;
  int                       i, j, result = 0;
  cardinfo*                 boardinfo = (cardinfo*)flip->private_data;
  unsigned long             ptsize;
  DWORD*                    ppt;
  int pagecounter = 0;

  DBGM("Entering MMap\n");



  //create memory area for bmp page table
  ptbmpuser = (DWORD)pci_alloc_consistent(boardinfo->pdev, 4000,&ptbmpphys);

  //create memory area for bmp RPS code
  rpsbmpuser = (DWORD)pci_alloc_consistent(boardinfo->pdev, 4000,&rpsbmpphys);


  //create memory area for jpg page table
  ptjpguser = (DWORD)pci_alloc_consistent(boardinfo->pdev, 4000,&ptjpgphys);


  //create memory area for jpg RPS code
  rpsjpguser = (DWORD)pci_alloc_consistent(boardinfo->pdev, 4000,&rpsjpgphys);



  //create memory area for entire buffer
  start = vma->vm_start;
  size = vma->vm_end - vma->vm_start;
  vma->vm_flags |= VM_LOCKED;

 for(j=0;j<SYS_BUFFERS;j++)
    {
      if(boardinfo->bufferinfo[j].jpegbufferuser==0)
	 break;
    }

  boardinfo->bufferinfo[j].jpegbufferuser = start;
  DBGM("Buffer # %i = %lx number of pages size %lx = \n", j, start, size);

  ppt = (DWORD*)ptjpguser;
  ptsize = 0;
  for (i = 0; i < (size / PAGE_SIZE); i++)
    {

      pageaddy = get_free_page (GFP_KERNEL);
      if (pageaddy == 0)
	{
	  printk ("get_free_page failed for buffer\n");
	  return -ENOBUFS;
	}
      paddress = virt_to_phys((void*)pageaddy);
      mem_map_reserve (virt_to_page (pageaddy));
      result =
	remap_page_range (start, paddress, PAGE_SIZE,
			  vma->vm_page_prot);


      //the first SYS_JPGSIZE bytes are the jpg buffer
      //the next SYS_BMPSIZE are the bmp buffer

      if(ptsize <SYS_JPGSIZE)
	{
	  ptsize += PAGE_SIZE;
	  *ppt++ = paddress;
	  DBGM("Added page %lx to jpg pt #%X ptsize = %lx\n", paddress, ++pagecounter, ptsize);
	}
      else

	if(ptsize == SYS_JPGSIZE)
	  { 
	    ptsize += PAGE_SIZE;
	    ppt = (DWORD*)ptbmpuser;
	    pagecounter=0;
	    *ppt++ = paddress;
	    boardinfo->bufferinfo[j].bmpbufferuser = start;
	    DBGM("Added page %lx to bmp pt#%X ptsize = %lx\n", paddress, ++pagecounter, ptsize);
	  }
	else
	  if((ptsize > SYS_JPGSIZE) && (ptsize < (SYS_JPGSIZE + SYS_BMPSIZE)))
	    {
	      ptsize += PAGE_SIZE;
	      *ppt++ = paddress;
	      DBGM("Added page %lx to bmp pt#%X ptsize = %lx\n", paddress, ++pagecounter, ptsize);
	    }

      if (result)
	return -EAGAIN;
      start += PAGE_SIZE;
      addpagetolist(boardinfo->ppagelist, paddress);

    }
  boardinfo->bufferinfo[j].ptbmpphys = ptbmpphys;
  boardinfo->bufferinfo[j].ptbmpuser = ptbmpuser;
  boardinfo->bufferinfo[j].rpsbmpphys = rpsbmpphys;
  boardinfo->bufferinfo[j].ptjpgphys = ptjpgphys;
  boardinfo->bufferinfo[j].ptjpguser = ptjpguser;
  boardinfo->bufferinfo[j].rpsjpgphys = rpsjpgphys;
  boardinfo->bufferinfo[j].rpsbmpuser = rpsbmpuser;
  boardinfo->bufferinfo[j].rpsjpguser = rpsjpguser;
  printk("%p, %p, %p, %p, %lx, %lx\n",(void*)ptbmpphys,(void*)rpsbmpphys,(void*)ptjpgphys,(void*)rpsjpgphys,rpsbmpuser,rpsjpguser); 
  return 0;
}







static int
iocontrol (struct inode *inode, struct file *flip,
	   unsigned int cmd, unsigned long arg)
{
  int i;
  cardinfo* boardinfo = flip->private_data;
  //  printk("ioctl command %x ioctl argument %lx\n",cmd&0xFF, arg);

  //  waitabit();
  //  waitabit();
  switch (cmd)
    {
    case S615Start:
      SetMode(boardinfo);
      break;
    case S615Stop:
      DisableInterrupts(boardinfo);
   break;
    case S615FreeBuffer:
      get_user(i, (int*)arg);
      boardinfo->bufferinfo[i].complete = 0;
      boardinfo->bufferinfo[i].read = 0;
      boardinfo->bufferinfo[i].busy = 0;
      break;
    default:
      return -ENOTTY;
      break;
    }

  return 0;
}

static int InitCard(int minor)
{
  unsigned long io_addr;
  unsigned long io_size;
  struct pci_dev *pdev = board[minor].pdev;
  char print_name[sizeof (pdev->slot_name)];
  int ecode, i;
  BYTE bdata;

  DBG("InitCard(%d)\n", minor);
  io_addr = pci_resource_start (pdev, 0);
  io_size = pci_resource_len (pdev, 0);
  memcpy (print_name, pdev->slot_name, sizeof (pdev->slot_name));
  if (!request_mem_region (io_addr, io_size, print_name))
    {
      DBG ("Memory Region request failed\n");
      return -EIO;
    }
  board[minor].base = ioremap (io_addr, io_size);
  if (board[minor].base == 0)
    {
      DBG ("IOREMAP failed\n");
      return -EIO;
    }
  WR7146 (minor, P_MC1, SOFT_RESET);	//soft reset 7146
  WR7146 (minor, P_MC1, I2C_ENABLE | DEBI_ENABLE);	//enable I2C and DEBI
  WR7146 (minor, P_INTEN, INT_OFF);	//disable interrupts
  WR7146 (minor, P_GPIO, GPIO_RST);	//init GPIO's

  //Init I2C
  ecode = InitI2C (minor);
  if (ecode)
    {
      return ecode;
    }
  //Init 7146
  Init7146 (minor);

  //Init GA pins
  WriteGA (minor, R6, R6_OUTPINEN);
  //Set FRAME to 1
  bdata = ReadGA (minor, R4);
  bdata |= 4;
  WriteGA (minor, R4, bdata);

  ecode = InitXP (minor);
  if (ecode)
    {
      return ecode;
    }
    
  ecode = LoadSRAM (minor);
  if (ecode)
    {
      return ecode;
    }
    

  //Load color constants
  for (i = 0; i < 8; i++)
    {
      WriteGA (minor, LOBYTE (LOWORD ((8 + i) << 1)), CLR[i]);
    }

  board[minor].xpdata = XP_INIT;

  ecode = S615_RequestIRQ(&board[minor]);
  if(ecode)
    {
      return ecode;
    }
  board[minor].dointerrupts = TRUE;

  board[minor].mode.format = FORMAT_NTSC;
  board[minor].mode.bmpflip = FLIP_ON;
  board[minor].mode.bmpcolor =COLOR_YCRCB_BS; 
  board[minor].mode.bmpsize = BMP_SIZE_LRG;
  board[minor].mode.jpgsize = JPG_SIZE_FULL;
  board[minor].mode.captionformat = CAP16X4;
  board[minor].mode.captionpos = CP_UL;
  board[minor].mode.fontsize = CF_LARGE;
  board[minor].mode.compfactor = 0xe0;// COMPFACTOR_MIN;
  board[minor].mode.xpinit = XPS_INIT;

  board[minor].cardstatus.jpegon = JPEG_ON;
  board[minor].cardstatus.captionmode = CAP_ON | CAP_MGT;	
  board[minor].cardstatus.interlacemode = INT_INT;	//interlaced mode



  return ecode;
}



static int __devinit
S615_init_one (struct pci_dev *pdev, const struct pci_device_id *ent)
{
  /* wake up and enable device */
  if (pci_enable_device (pdev))
  {
      return -EIO;
  } else {
      board[S615_cards_found].pdev = kmalloc (sizeof (*pdev), GFP_ATOMIC);
      memcpy (board[S615_cards_found].pdev, pdev, sizeof (*pdev));
      DBG ("Board %d found\n",S615_cards_found);
  }

  S615_cards_found++;

  return 0;

}

static void __devexit
S615_remove_one (struct pci_dev *pdev)
{
  int i;
  int remove = -1;
  DBG ("Entering S615_remove_one\n");
  for (i = 0; i < SYS_BOARDS; i++)
    if (board[i].pdev)
      {
	DBG ("request to remove board in slot %s current board slot %s %d\n",
	     pdev->slot_name, board[i].pdev->slot_name, i);
	if (strcmp (pdev->slot_name, board[i].pdev->slot_name) == 0)
	  {
	    DBG ("MATCH!!!\n");
	    remove = i;
	    break;
	  }
      }
  if (remove == -1)
    {
      DBG ("BAD REQUEST for Card Removal Slot %s\n", pdev->slot_name);
      return;
    }
  S615_cards_found--;
  DBG ("remove=%d S615_dev=0x%x\n", remove,
       (unsigned int) board[remove].pdev);
  kfree (board[remove].pdev);
  board[remove].pdev = 0;
};

static struct pci_driver S615_driver = {
  name:"S615",
  probe:S615_init_one,
  remove:S615_remove_one,
  id_table:S615_id_tbl,
};



int __init
S615_init (void)
{

  int i;
  //  struct pagelist* list = kmalloc(sizeof(struct pagelist), GFP_KERNEL);
  DBG ("Entering S615 Module Init\n");
  for (i = 0; i < SYS_BOARDS; i++)
    {
      board[i].pdev = 0;
      init_waitqueue_head (&board[i].wait_queue);
      board[i].wait_condition = 0;
      board[i].dointerrupts = FALSE;
      board[i].interruptcounter = 0;
      board[i].interruptslost = 0;
    }

  devicemajornumber = register_chrdev (OURMAJOR, driverdevice, &S615_ops);
  if(devicemajornumber != 0)
    {
      printk(KERN_WARNING " chosen major number of %i is not available, revise h file and recompile\n", OURMAJOR);
      return -ENODEV;
    }
  devicemajornumber = OURMAJOR;

	// following is for Intel i845 chipset bug fix
	// finding Intel i845 chipset in the system. if it's in, make fix.
	dev_i845 = pci_find_device ( i845DB_pci_tbl[0].vendor, i845DB_pci_tbl[0].device, dev_i845 );
	if ( dev_i845 != NULL ) {
	   pci_read_config_dword ( dev_i845, 0x50, &i845_ConfigSpace_0x50 );
	   //printk ("S615: found i845 DB.  i845_ConfigSpace_0x50 = 0x%x \n", i845_ConfigSpace_0x50 );
	   i845_ConfigSpace_0x50 |= 0x4;
	   pci_write_config_dword ( dev_i845, 0x50, i845_ConfigSpace_0x50 );
	   //pci_read_config_dword ( dev_i845, 0x50, &i845_ConfigSpace_0x50 );
	   //printk ("S615: after fix.      i845_ConfigSpace_0x50 = 0x%x \n", i845_ConfigSpace_0x50 );
	}
	dev_i845 = pci_find_device ( i845BA_pci_tbl[0].vendor, i845BA_pci_tbl[0].device, dev_i845 );
	if ( dev_i845 != NULL ) {
	   pci_read_config_dword ( dev_i845, 0x50, &i845_ConfigSpace_0x50 );
	   //printk ("S615: found i845 BA.  i845_ConfigSpace_0x50 = 0x%x \n", i845_ConfigSpace_0x50 );
	   i845_ConfigSpace_0x50 |= 0x4;
	   pci_write_config_dword ( dev_i845, 0x50, i845_ConfigSpace_0x50 );
	   //pci_read_config_dword ( dev_i845, 0x50, &i845_ConfigSpace_0x50 );
	   //printk ("S615: after fix.      i845_ConfigSpace_0x50 = 0x%x \n", i845_ConfigSpace_0x50 );
	}
	// done fix

  printk("S615: %s installed - device major number = %d\n", driverdevice, devicemajornumber);
  return pci_module_init (&S615_driver);
}


void __exit
S615_cleanup (void)
{
  DBG ("Entering S615 Module Cleanup\n");
  unregister_chrdev (devicemajornumber, driverdevice);
  pci_unregister_driver (&S615_driver);
}

static DWORD GetFreeBuffer(cardinfo* boardinfo)
{
  int i;
  for(i=0;i<SYS_BUFFERS;i++)
    if(!boardinfo->bufferinfo[i].busy)
      return i;
  return SYS_NOFREEBUFFERS;
}

static int GetAvailableBuffer(cardinfo* boardinfo)
{
  int i;
  for(i=0;i<SYS_BUFFERS;i++)
    if(boardinfo->bufferinfo[i].busy && boardinfo->bufferinfo[i].complete && !boardinfo->bufferinfo[i].read)
      {
	boardinfo->bufferinfo[i].read = 1;
	return i;
      }
  return SYS_NOFREEBUFFERS;
}


static unsigned long get_time(void)
{
  struct timeval tv;
  do_gettimeofday(&tv);
  return (1000000 * tv.tv_sec + tv.tv_usec);
}

static void
handleallinterrupts (int irq, void *dev_id, struct pt_regs *regs)
{
  int index;
  DWORD irqstatus;
  cardinfo *boardinfo = dev_id;
  BOOL   reg;
  static BOOL wait = FALSE, started = FALSE;
  static int  bufnum = 0;
  static BYTE i2cdata, bdata;
  BYTE   dcdraddr[2] = { D_I2CADDR0, D_I2CADDR1 };
  char cap[64] = "\0";
  struct timeval ts;

  
  if (dev_id == 0)
    {
      DBGI("Bad Boy: Got interrupt # %d with a NULL dev_id\n", irq);
      goto end;
    }
  for (index = 0; index < SYS_BOARDS;index++)
    {
      if ( &board[index] == boardinfo)
	{
	  //The first thing to do is to determine if this S615 created this interrupt.
	  irqstatus = readl (boardinfo->base + P_INTSTAT);      
	  if ((irqstatus & IRQ_MASK) == 0)
	    {
	      // this is not our interrupt
	      DBGI ("irqstatus %lx board base %p board number%i\n",
		    irqstatus, boardinfo->base, index);
	      goto end;
	    }
	  //This is our interrupt so clear the board's master interrupt enable bit.
	  DisableInterrupts(boardinfo);
	  //increment the counters
	  boardinfo->interruptcounter++;
	  //	  boardinfo->interruptslost++;
	  DBGI("getting interrupt# 0x%lx missed 0x%lx irqstatus 0x%lx index %i\n", 
	       boardinfo->interruptcounter, boardinfo->interruptslost, irqstatus, index);

	  WR7146 (index, P_GPIO, 0x00205040);	//~~set GPIO0 to 0

	  if (started)
	    {
	      if (RR7146 (index, P_MC2) & STATUS_MC2_SIG1)//if RPS1 not done
		{			
		  if (RR7146 (index, P_PSR) & PSR_PIN3)
		    {
		      ++boardinfo->cardstatus.incomplete;	//error: Zoran not done by next VSYNC
		      if (wait)
			{			//we've already waited for 1 frame
			  WR7146 (index, P_MC1, DSBL_RPS | DSBL_DEBI);	////disable RPS1,2,video1+DEBI
			  WR7146 (index, P_MC2, STATUS_MC2_SIG1 << 16);	//reset SIG1
			  wait = FALSE;
			}
		      else
			{
			  wait = TRUE;
			}
		    }
		  else
		    {
		      WR7146 (index, P_MC1, DSBL_RPS | DSBL_DEBI);	////disable RPS1,2,video1+DEBI
		      WR7146 (index, P_MC2, STATUS_MC2_SIG1 << 16);	//reset SIG1
		    }
		}
	      else
		{
		  wait = FALSE;
		}

	      if (!wait)
		{
		  //this has to be done before Zoran reg access
		  boardinfo->bufferinfo[bufnum].jpegsize = RR7146 (index, P_DEBIAD);	//JPEG capture was enabled, store Jpeg size
		  bdata = RR060 (index, C_IRQSTAT);
		  if (bdata & 1)//DATERR#
		    {
		      boardinfo->bufferinfo[bufnum].jpegsize = 0;
		      ++boardinfo->cardstatus.zerror;
		    }
		  else if (boardinfo->cardstatus.jpegon)//END# low: complete or was not started
		    {			
		      if (boardinfo->bufferinfo[bufnum].jpegsize > SYS_JPGSIZE)
			{
			  boardinfo->bufferinfo[bufnum].jpegsize = 0;
			}
		    }
		  else  //JPEG capture was not enabled
		    {			
		      boardinfo->bufferinfo[bufnum].jpegsize = 0;
		    }
		  boardinfo->bufferinfo[bufnum].camera = GetXP (boardinfo,boardinfo->bufferinfo[bufnum].channel);	//store xp channel #
		  //printk("\ti2cdata = 0x%x\n", i2cdata);
		  boardinfo->bufferinfo[bufnum].videopresent = (i2cdata&0x40)==0?1:0;
		  //boardinfo->bufferinfo[bufnum].videopresent = i2cdata & 1;

		  //we have an image here!!!
		  //we have an image here!!!
		  //we have an image here!!!
		  
		  boardinfo->bufferinfo[bufnum].complete = 1;


		  //we have an image here!!!
		  //we have an image here!!!
		  //we have an image here!!!

		  SetXP (index, LOBYTE (LOWORD (boardinfo->bufferinfo[bufnum].channel)),
			 LOBYTE (LOWORD (boardinfo->bufferinfo[bufnum].xpchan)));

		  started = FALSE;
		}
	    }

	  if(!wait)		  
	    {
	      //setup command structure here as we are about to start another grab
	      bufnum = GetFreeBuffer (boardinfo);

	      if (bufnum < SYS_NOFREEBUFFERS)
		{			//if free buffer found
		  boardinfo->bufferinfo[bufnum].complete = 0;	
		  boardinfo->bufferinfo[bufnum].read = 0;	
		  boardinfo->bufferinfo[bufnum].busy = 1;	//mark buffer as busy
	          GetChanInp(boardinfo, &boardinfo->bufferinfo[bufnum].channel, 
			     &boardinfo->bufferinfo[bufnum].xpchan);	
		  printk("GotNext decoder = %i, camera = %i\n",boardinfo->bufferinfo[bufnum].channel,boardinfo->bufferinfo[bufnum].xpchan); 
	  //CaptionFE = FIFOread (&g_CapFIFO[index]);

		  //Read decoder status bit
		  //switch I2C mux
   
		  bdata = ReadGA (index, R6);
		  if (CHAN[boardinfo->bufferinfo[bufnum].channel & 3] <= 1)
		    {
		      bdata &= 0xfe;	//clear bit 0
		    }
		  else
		    {
		      bdata |= 1;	//set bit 0
		    }
		  WriteGA (index, R6, bdata);
		  ReadI2C (index, dcdraddr[CHAN[boardinfo->bufferinfo[bufnum].channel & 3] % 2], 0x1f,
			   &i2cdata);

		  bdata =
		    CHAN[boardinfo->bufferinfo[bufnum].channel & 3] | 
		    (boardinfo->cardstatus.captionmode & 1) << 3 | 
		    (boardinfo->cardstatus.captionmode & 0x70)<< 1;

		  WriteGA (index, R7, bdata);	//control word (R7)
		  bdata = ReadGA (index, R4);
		  bdata &= 0x37;	//clear intrlce mode and START
		  bdata |= LOBYTE (LOWORD (boardinfo->cardstatus.interlacemode & 3)) << 6;//set interlace mode
		  WriteGA (index, R4, bdata);	//write intrlc mode

		  if (boardinfo->cardstatus.jpegon)
		    {
		      bdata |= 8;
		      WriteGA (index, R4, bdata);	//write START
		      do_gettimeofday(&ts);
		    


		      boardinfo->bufferinfo[bufnum].camera = 
			GetXP (boardinfo, boardinfo->bufferinfo[bufnum].channel);
		      //caption stuff here
		      sprintf(cap, "testcaption     %lu ", get_time());
		      WriteCaption(boardinfo,boardinfo->bufferinfo[bufnum].channel, cap); 


		      {
			udelay (15);	//15us delay
		      }
		      bdata = ReadGA (index, R4);
		      bdata &= ~8;	//clear START bit
		      WriteGA (index, R4, bdata);
		    }
		  StartRPS (index, &boardinfo->bufferinfo[bufnum], boardinfo->cardstatus.jpegon);
		  started = TRUE;
		}
	      else
		{
		  ++boardinfo->interruptslost;
		  ++boardinfo->cardstatus.bqfull;

		}
	  
	    }
	  reg = TRUE;
	  while (reg)
	    {
	      WR7146 (index, P_INTEN, INTSTATUS_VSYNC);
	      reg = (RR7146 (index, P_INTEN) != INTSTATUS_VSYNC);
	    }

	  WR7146 (index, P_GPIO, 0x00205050);	//~~set GPIO0 to 1

	

	}

      goto end;
    }

 end:
  return;
}


/*  		      TranslateCaption (index, CaptionOut, CaptionFE.caption, */
/*  					GetXP (boardinfo, boardinfo->cardstatus.channel), */
/*  					boardinfo->cardstatus.offset); */
/*  		    if (boardinfo->cardstatus.captionmode & 1) */
/*  		      {		//if caption ON */
/*  			WriteCaption (index, boardinfo->cardstatus.channel, CaptionOut);//write caption */
/*  		      } */
/*  		    else */





static int
S615_RequestIRQ (cardinfo * board)
{

  unsigned int irq = board->pdev->irq;
  void *dev_id = board;		//pointer to the boards information
  unsigned long flags = SA_SHIRQ;
  int request = -1;
  //disable the interrupt on this board before we turn on the handler
  //  writel (0, board->base + P_IER);
  //  writel (IRQ_MASK, board->base + P_ISR);
  DisableInterrupts(board);
  request =
    request_irq (irq, handleallinterrupts, flags, board->name, dev_id);
  if (request)
    {
      DBG ("did not get interrupt %d\n", irq);
      //failed


    }
  else
    {
      DBG ("successfully requested handler for irq %d id = %p\n", irq,
	      dev_id);
    }
  return request;
}

static void Init7146 (int index)
{

  WR7146 (index, P_D1SYNC, D1_INIT);	//0x050; Init 7146 D1 interface
  WR7146 (index, P_D1VIDEO, 0);	//0x054; ~~may req mod for PAL (see 613)
  WR7146 (index, P_PCIBTV, BURST_INIT);	//0x048; burst, threshold
  WR7146 (index, P_HPSCTL, HPS_INIT);	//0x05c;
  WR7146 (index, P_BCSCTL, BCS_INIT);	//0x070;
  WR7146 (index, P_CKEY, 0);

  WR7146 (index, P_MC1, VDP_DEBI_ENABLE);

  WR7146 (index, P_INTSTAT, 0xffffffff);	//clear status
  WR7146 (index, P_MC2, 0xf8000000);	//clear sigs
  return;
}

static ECODE InitI2C (DWORD index)
{
  DWORD                     data;
  int                       i;

  for (i = 0; i < 3; i++)
    {
      //...ABORT (to clear BUSY)
      WR7146 (index, P_I2CSTAT, I2C_STATSTUB | I2C_ABORT);
      WR7146 (index, P_MC2, UPLD_IIC);	//upload;
      //              xdelay(500);
      while (!(RR7146 (index, P_MC2) & UPLD_IIC_MASK));

      //...clear STATUS reg twice
      WR7146 (index, P_I2CSTAT, I2C_STATSTUB);
      WR7146 (index, P_MC2, UPLD_IIC);	//upload;
      //              xdelay(500);
      while (!(RR7146 (index, P_MC2) & UPLD_IIC_MASK));

      WR7146 (index, P_I2CSTAT, I2C_STATSTUB);
      WR7146 (index, P_MC2, UPLD_IIC);	//upload;
      //              xdelay(500);
      while (!(RR7146 (index, P_MC2) & UPLD_IIC_MASK));

      //Check BUSY and ERROR
      data = RR7146 (index, P_I2CSTAT);
      if (((data & I2C_BUSY) == 0) && ((data & I2C_ERR) == 0))
	{
	  break;
	}
    }
  if (data & I2C_BUSY)
    {
      return ERR_I2CINIT_BUSY;
    }
  else if (data & I2C_ERR)
    {
      return ERR_I2CINIT_ERR;
    }
  else
    {
      return 0;
    }
  return 0;
}

static void WriteGA (int index, WORD addr, BYTE data)
{
  DWORD                     dwdata;

  //Set up DEBI registers
  WR7146 (index, P_DEBICFG, DEBI_CFG_IMMEADACCESS);
  WR7146 (index, P_DEBIPAGE, DEBI_PAGE_IMMEADACCESS);

  dwdata = DEBI_CMD_WR1 | addr;
  WR7146 (index, P_DEBICMD, dwdata);
  //set up AD register
  WR7146 (index, P_DEBIAD, data);
  WR7146 (index, P_MC2, UPLD_DEBI);	//upload;
  WaitDEBIDone (index);
  return;
}

static BYTE ReadGA (int index, WORD addr)
{
  DWORD                     dwdata;

//      BYTE bdata;

  //Set up DEBI registers
  WR7146 (index, P_DEBICFG, DEBI_CFG_IMMEADACCESS);
  WR7146 (index, P_DEBIPAGE, DEBI_PAGE_IMMEADACCESS);

  dwdata = DEBI_CMD_RD1 | addr;
  WR7146 (index, P_DEBICMD, dwdata);
  //clear AD register
  WR7146 (index, P_DEBIAD, 0);
  WR7146 (index, P_MC2, UPLD_DEBI);	//upload;

  WaitDEBIDone (index);

  dwdata = RR7146 (index, P_DEBIAD);
  return LOBYTE (LOWORD (dwdata));
}


static void WR060 (int index, WORD addr, BYTE bdata)
{
  WriteGA (index, A_ZORAN | Z_ADDR_LSB, LOBYTE (addr));
  WriteGA (index, A_ZORAN | Z_ADDR_MSB, HIBYTE (addr));
  WriteGA (index, A_ZORAN | Z_DATA, bdata);

  return;
}

static BYTE RR060 (int index, WORD addr)
{
  BYTE                      bdata;

  WriteGA (index, A_ZORAN | Z_ADDR_LSB, LOBYTE (addr));
  WriteGA (index, A_ZORAN | Z_ADDR_MSB, HIBYTE (addr));
  bdata = ReadGA (index, A_ZORAN | Z_DATA);

  return bdata;
}

//~~
WORD RC060 (int index)
{
  //Reads code FIFO of 060
  DWORD                     dwdata;

  //Set up DEBI registers
  WR7146 (index, P_DEBICFG, DEBI_CFG_IMMEADACCESS | DEBI_CFG_16);
  WR7146 (index, P_DEBIPAGE, DEBI_PAGE_IMMEADACCESS);

  dwdata = DEBI_CMD_RD2 | A_ZORAN;
  WR7146 (index, P_DEBICMD, dwdata);
  //clear AD register
  WR7146 (index, P_DEBIAD, 0);
  WR7146 (index, P_MC2, UPLD_DEBI);	//upload;

  WaitDEBIDone (index);

  return LOWORD (RR7146 (index, P_DEBIAD));
}

//~~

static ECODE Init060 (int index, MODE * pmode)
{
  ECODE                     ecode = 0;

  Reset060 (index);
  ecode = Load060 (index, pmode);

  return ecode;
}

static void Reset060 (int index)
{

  WriteGA (index, A_GA | R4, R4_ZSLEEP);	//Zoran RESET low
  udelay (150);
  WriteGA (index, A_GA | R4, R4_ZRESET | R4_ZSLEEP);	//Zoran RESET and SLEEP high
  udelay (150);
  WriteGA (index, A_GA | R4, R4_ZRESET);	//Zoran SLEEP low
  udelay (150);
  WriteGA (index, A_GA | R4, R4_ZRESET | R4_ZSLEEP);	//Zoran SLEEP high
  udelay (300);
  return;
}

static ECODE Load060 (int index, MODE * pmode)
{
  ECODE                     ecode = 00;
  WORD                      i, addr;
  BYTE                      regval;
  BOOL                      unclipped;

  //Load base values
  if (pmode->format == FORMAT_NTSC)
  {
    for (i = 0; i < 0x400; i++)
    {
      WR060 (index, i, REG_C_NTSC[i]);
    }
  }
  else
  {
    for (i = 0; i < 0x400; i++)
    {
      WR060 (index, i, REG_C_PAL[i]);
    }
  }

  unclipped = ((pmode->jpgsize == JPG_SIZE_FULL)
	       || (pmode->jpgsize == JPG_SIZE_HALF));

  //Load values dependent on MODE settings
  if (pmode->format == FORMAT_NTSC)
  {
    if (unclipped)
    {				//non-clipped
      for (i = 0, addr = Z_SYNC_START; i < Z_SYNC_BYTES; i++, addr++)
      {
	WR060 (index, addr, SYNC_NTSC[0][i]);
      }
    }
    else
    {				//clipped
      for (i = 0, addr = Z_SYNC_START; i < Z_SYNC_BYTES; i++, addr++)
      {
	WR060 (index, addr, SYNC_NTSC[1][i]);
      }
    }
    if ((pmode->jpgsize == JPG_SIZE_FULL)
	|| (pmode->jpgsize == JPG_SIZE_FULL_C))
    {				//no scaling
      WR060 (index, Z_SCALE_START, Z_SCALE1);
    }
    else
    {
      WR060 (index, Z_SCALE_START, Z_SCALE2);
    }
    for (i = 0, addr = Z_SUBIMG_START; i < Z_SUBIMG_BYTES; i++, addr++)
    {
      WR060 (index, addr,
	     SUBIMG_NTSC[unclipped ? 0 : 1][pmode->fontsize][pmode->
							     captionpos]
	     [pmode->captionformat][i]);
    }
    for (i = 0, addr = Z_VSIZE_START; i < Z_VSIZE_BYTES; i++, addr++)
    {
      WR060 (index, addr, VSIZE_NTSC[i]);
    }
    for (i = 0, addr = Z_MRK_X; i < Z_MRK_BYTES; i++, addr++)
    {
      WR060 (index, addr, JPEG_MRK_NTSC[pmode->jpgsize][i]);
    }
  }
  else
  {
    if (unclipped)
    {				//non-clipped
      for (i = 0, addr = Z_SYNC_START; i < Z_SYNC_BYTES; i++, addr++)
      {
	WR060 (index, addr, SYNC_PAL[0][i]);
      }
    }
    else
    {				//clipped
//                      for (i = 0, addr = Z_SYNC_START; i < Z_SYNC_BYTES; i++, addr++) {
//                              WR060 (index, addr, SYNC_PAL[1][i]);
//                      }
    }
    if ((pmode->jpgsize == JPG_SIZE_FULL)
	|| (pmode->jpgsize == JPG_SIZE_FULL_C))
    {				//no scaling
      WR060 (index, Z_SCALE_START, Z_SCALE1);
    }
    else
    {
      WR060 (index, Z_SCALE_START, Z_SCALE2);
    }
    for (i = 0, addr = Z_SUBIMG_START; i < Z_SUBIMG_BYTES; i++, addr++)
    {
      WR060 (index, addr,
	     SUBIMG_PAL[unclipped ? 0 : 1][pmode->fontsize][pmode->
							    captionpos]
	     [pmode->captionformat][i]);
    }
    for (i = 0, addr = Z_VSIZE_START; i < Z_VSIZE_BYTES; i++, addr++)
    {
      WR060 (index, addr, VSIZE_PAL[i]);
    }
    for (i = 0, addr = Z_MRK_X; i < Z_MRK_BYTES; i++, addr++)
    {
      WR060 (index, addr, JPEG_MRK_PAL[pmode->jpgsize][i]);
    }
  }

  WR060 (index, C_SCALEF15, HIBYTE (LOWORD (pmode->compfactor)));
  WR060 (index, C_SCALEF7, LOBYTE (LOWORD (pmode->compfactor)));
  WR060 (index, C_LOAD, CC_LOADBUSY);	//set LOAD bit (start loading parameters)

  //wait for load to complete
  i = 0;
  regval = CC_LOADBUSY;
  while ((regval & CC_LOADBUSY) && (i < CC_TIMEOUT))
  {
    udelay (15);
    regval = RR060 (index, C_CFSTAT);
    i++;
  }

  WR060 (index, C_LOAD, 0);	//reset LOAD bit;
  if (i == CC_TIMEOUT)
  {
    ecode = ERR_TIMEOUT;
  }

  //reset sync gen
  WR060 (index, C_LOAD, CC_RESETSYNC);
  udelay (15);
  WR060 (index, C_LOAD, CC_STARTSYNC);
  return ecode;
}
static ECODE InitXP (int index)
{
  ECODE                     ecode;

  if ((ecode = LoadXP (index, XP_PREINIT)))
  {
    return ecode;
  }
  else if ((ecode = LoadXP (index, XP_INIT)))
  {
    return ecode;
  }
  else
  {
    return 0;
  }
}

static void WriteSRAM (int index, WORD addr, BYTE data)
{
  WriteGA (index, (WORD) (A_SRAM | addr), data);
  return;
}

static BYTE ReadSRAM (int index, WORD addr)
{
  return ReadGA (index, (WORD) (A_SRAM | addr));
}

static ECODE LoadSRAM (int index)
{
  int                       i;
  BYTE                     *pFonts = (BYTE *) fnt;
  BYTE                      bdata;
  ECODE                     ecode = 0;

  for (i = 0; i < 0x1000; i++)
  {
    WriteSRAM (index, LOWORD (i), *pFonts++);
  }

  pFonts = (BYTE *) fnt;
  for (i = 0; i < 0x1000; i++)
  {
    bdata = ReadSRAM (index, LOWORD (i));
    if (bdata != *pFonts++)
    {
      ecode = ERR_SRAM;
      break;
    }
  }
  return ecode;
}

static void WaitDEBIDone (int index)
{
  //Reads DEBI Status bit from PSR. If it goes from 1 to 0 - quits
  //otherwise quits after 5 tries.
  int                       i;
  BOOL                      one = FALSE;

  for (i = 0; i < DEBI_DONE_WAIT; i++)
  {
    if (RR7146 (index, P_PSR) & DEBI_S)
    {
      one = TRUE;
    }
    else if (one)
    {
      break;
    }
  }
  return;
}

static ECODE LoadXP (int index, DWORD xpdata)
{
  BYTE                      data;

  //Load R0-R3
  data = HIBYTE (HIWORD (xpdata));
  WriteGA (index, R0, data);
  data = LOBYTE (HIWORD (xpdata));
  WriteGA (index, R1, data);
  data = HIBYTE (LOWORD (xpdata));
  WriteGA (index, R2, data);
  data = LOBYTE (LOWORD (xpdata));
  WriteGA (index, R3, data);

  //Start state machine: write 1 to R5
  WriteGA (index, R5, R5_START_XPSM);

  udelay (75);
  data = ReadGA (index, R5);
  if (data & R5_START_XPSM)
  {
    return ERR_XPTO;
  }
  return 0;

}

static ECODE Init7113 (int index, int decno, MODE * pmode)
{
  ECODE                     ecode = 0;
  BYTE                      bdata, addr;
  int                       i;

  //switch I2C mux
  bdata = ReadGA (index, R6);
  if (decno <= 1)
  {
    bdata &= 0xfe;		//clear bit 0
  }
  else
  {
    bdata |= 1;			//set bit 0
  }
  WriteGA (index, R6, bdata);

  if (decno % 2)
  {
    addr = D_I2CADDR1;
  }
  else
  {
    addr = D_I2CADDR0;
  }

  if (pmode->format == FORMAT_NTSC)
  {
    for (i = 0; i < D_REGS; i++)
    {
      if ((ecode =
	   WriteI2C (index, addr, REGS_D_NTSC[i][0], REGS_D_NTSC[i][1])))
      {
	break;
      }
    }
  }
  else if (pmode->format == FORMAT_PAL)
  {
    for (i = 0; i < D_REGS; i++)
    {
      if ((ecode =
	   WriteI2C (index, addr, REGS_D_PAL[i][0], REGS_D_PAL[i][1])))
      {
	break;
      }
    }
  }
  else
  {
    ecode = ERR_ILLEGAL_MODE;
  }

  return ecode;
}

static ECODE SetMode (cardinfo* boardinfo)
{
  ECODE                     ecode = 0;
  int                       i;
  BYTE                      bdata;
  DWORD                     lines, pixels, bytesperpix, colorformat;
  DWORD                     hpshps, hpshfs, hpsvs, hpsvsg;
  MODE*                     pmode = &(boardinfo->mode);
  int                       index = boardinfo->myminor;



    WR7146 (index, P_INTEN, INT_OFF);	//disable interrupts
    WR7146 (index, P_MC1, DSBL_RPS);	//disable RPS's

    if (pmode->compfactor < COMPFACTOR_MIN)
    {
      pmode->compfactor = COMPFACTOR_MIN;
    }
    if (!(ecode = Init060 (index, pmode)))
    {
      for (i = 0; i < 4; i++)
      {
	if ((ecode = Init7113 (index, i, pmode)))
	{
	  break;
	}
      }
    }

    //Set GA registers depending on MODE
    bdata = ReadGA (index, R6);
    bdata &= 0xE3;
    bdata |=
      ((LOBYTE (LOWORD (pmode->captionformat)) & 3) << 2) |
      ((~pmode->fontsize & 1) << 4);
    WriteGA (index, R6, bdata);
    //Restore FRAME# to 1
    bdata = ReadGA (index, R4);
    bdata |= 4;
    WriteGA (index, R4, bdata);

    for (i = 0; i < SYS_BUFFERS; i++)
      {
	CreateRPS (&boardinfo->bufferinfo[i], pmode);
	boardinfo->bufferinfo[i].lpbmi = (LPBITMAPINFO) kmalloc (sizeof (BITMAPINFO) + 
							       255 * sizeof (RGBQUAD), GFP_KERNEL);
	CreateDIBHeader (&boardinfo->bufferinfo[i], pmode);
      }
    board[index].captionformat = pmode->captionformat;

    //Set 7146 reg's dependent on mode
    if (pmode->format == FORMAT_NTSC)
    {
      lines =
	(pmode->bmpsize ==
	 BMP_SIZE_LRG) ? BMPLINES_NTSC_LRG : BMPLINES_NTSC_SML;
      pixels =
	(pmode->bmpsize ==
	 BMP_SIZE_LRG) ? BMPPIXELS_NTSC_LRG : BMPPIXELS_NTSC_SML;
    }
    else
    {
      lines =
	(pmode->bmpsize ==
	 BMP_SIZE_LRG) ? BMPLINES_PAL_LRG : BMPLINES_PAL_SML;
      pixels =
	(pmode->bmpsize ==
	 BMP_SIZE_LRG) ? BMPPIXELS_PAL_LRG : BMPPIXELS_PAL_SML;
    }

    if ((pmode->bmpcolor == COLOR_YCRCB)
	|| (pmode->bmpcolor == COLOR_YCRCB_BS))
    {
      bytesperpix = BYTESPERPIX_YUV;
      colorformat = HPSFORM_INIT_YCRCB;
    }
    else if (pmode->bmpcolor == COLOR_RGB)
    {
      bytesperpix = BYTESPERPIX_RGB;
      colorformat = HPSFORM_INIT_RGB;	//RGB
    }
    else
    {
      bytesperpix = BYTESPERPIX_MONO;
      colorformat = HPSFORM_INIT_MONO;
    }

    if (pmode->bmpsize == BMP_SIZE_LRG)
    {
      hpshps = HPSHPS_INIT_LRG;
      hpsvs =
	(pmode->format ==
	 FORMAT_NTSC) ? HPSVS_INIT_LRG_NTSC : HPSVS_INIT_LRG_PAL;
      hpsvsg = HPSVSG_INIT_LRG;
      hpshfs = HPSHFS_INIT_LRG;
    }
    else
    {
      hpshps = HPSHPS_INIT_SML;
      hpsvs =
	(pmode->format ==
	 FORMAT_NTSC) ? HPSVS_INIT_SML_NTSC : HPSVS_INIT_SML_PAL;
      hpsvsg = HPSVSG_INIT_SML;
      hpshfs = HPSHFS_INIT_SML;
    }

    WR7146 (index, P_PROTADDR1, lines * pixels * bytesperpix);
    WR7146 (index, P_PITCH1, pixels * bytesperpix);
    WR7146 (index, P_HPSFORM, colorformat);
    WR7146 (index, P_HPSHPS, hpshps);
    WR7146 (index, P_HPSHFS, hpshfs);
    WR7146 (index, P_HPSVS, hpsvs);
    WR7146 (index, P_HPSVSG, hpsvsg);

    WR7146 (index, P_MC2, UPLD_ALLVIDEO);

    //Set xpoint switch
    if (!(ecode = SetXP (index, 0, HIBYTE (HIWORD (pmode->xpinit)))))
    {
      if (!(ecode = SetXP (index, 1, LOBYTE (HIWORD (pmode->xpinit)))))
      {
	if (!(ecode = SetXP (index, 2, HIBYTE (LOWORD (pmode->xpinit)))))
	{
	  if (!(ecode = SetXP (index, 3, LOBYTE (LOWORD (pmode->xpinit)))))
	  {
	    printk("Enabling\n");
	    WR7146 (index, P_INTEN, INTSTATUS_VSYNC);	//re-enable interrupt
	  }
	}
      }
    }

  return ecode;
}

static ECODE ReadI2C (int index, BYTE devaddr, BYTE addr, BYTE * pdata)
{
  DWORD                     data;
  int                       i = 0, to = 40;

  //write address
  data =
    (devaddr << 24) | (addr << 16) | (I2C_ATTRSTART << 6) | (I2C_ATTRSTOP <<
							     4) | (I2C_ATTRNOP
								   << 2);
  WR7146 (index, P_I2CCTRL, data);
  WR7146 (index, P_MC2, UPLD_IIC);	//upload;

  //wait until ready or error
  do
  {
    udelay (15);
    ++i;
    data = RR7146 (index, P_I2CCTRL);
  }
  while (((data & I2C_BUSY) != 0) && ((data & I2C_ERR) == 0) && (i < to));
  if (data & I2C_ERR)
  {
    return ERR_I2C;
  }
  else if (i == to)
  {
    return ERR_I2CTO;
  }

  //read register
  data =
    ((devaddr | 1) << 24) | (I2C_ATTRSTART << 6) | (I2C_ATTRSTOP << 4) |
    (I2C_ATTRNOP << 2);
  WR7146 (index, P_I2CCTRL, data);
  WR7146 (index, P_MC2, UPLD_IIC);	//upload;

  i = 0;
  //wait until ready or error
  do
  {
    udelay (15);
    ++i;
    data = RR7146 (index, P_I2CCTRL);
  }
  while (((data & I2C_BUSY) != 0) && ((data & I2C_ERR) == 0) && (i < to));
  if (data & I2C_ERR)
  {
    return ERR_I2C;
  }
  else if (i == to)
  {
    return ERR_I2CTO;
  }

  //read data
  data = RR7146 (index, P_I2CCTRL);

  *pdata = LOBYTE (data >> 16);
  return 0;
}

static ECODE WriteI2C (int index, BYTE devaddr, BYTE addr, BYTE bdata)
{
  DWORD                     dwdata;
  int                       i = 0, to = 40;

  //write address
  dwdata =
    (devaddr << 24) | (addr << 16) | (bdata << 8) | (I2C_ATTRSTART << 6) |
    (I2C_ATTRCONT << 4) | (I2C_ATTRSTOP << 2);
  WR7146 (index, P_I2CCTRL, dwdata);
  WR7146 (index, P_MC2, UPLD_IIC);	//upload;

  //wait until ready or error
  do
  {
    udelay (15);
    ++i;
    dwdata = RR7146 (index, P_I2CCTRL);
  }
  while (((dwdata & I2C_BUSY) != 0) && ((dwdata & I2C_ERR) == 0) && (i < to));

  if (dwdata & I2C_ERR)
  {
    return ERR_I2C;
  }
  else if (i == to)
  {
    return ERR_I2CTO;
  }

  return 0;
}

static ECODE SetXP (int index, BYTE out, BYTE inp)
{
  //out = [0..3]; inp = [1..16]; inp=0 turns the output off
  //.xpdata contains output settings for 2 muxes:
  //MSB...[out0_1][out1_1][out2_1][out[3_1][out0_0][out1_0][out2_0][out[3_0]...LSB
  //where mux_1 controls inputs 9-16, mux_0 - inputs 1-8
  //and outX_Y is [0..7]
  //disabled channel contains 0xa
  //before enabling a channel, 9 has to be written to it.

  DWORD                     dwbuf;
  int                       shifts;
  ECODE                     ecode;
  WORD                      lwbuf, hwbuf;
  DWORD                     xpdbuf = board[index].xpdata;

  lwbuf = LOWORD (xpdbuf);
  hwbuf = HIWORD (xpdbuf);
  dwbuf = MAKELONG (lwbuf, hwbuf);

  printk("entering SetXP dwbuf = %lx out= %x inp= %x\n", dwbuf, out, inp);
  shifts = (3 - out) * 4;
  if (inp > 8)
    shifts += 16;


  if (((dwbuf >> shifts) & 0xf) == CHAN_OFF)
  {				//if the channel was disabled
    //depending on inp, set corresponding nibble to 9 (CHAN_ON)
    //turn off the corresponding nibble in the other half of control dword
    if (inp > 8)
    {
      //set corresponding nibble of hiword to CHAN_ON, loword to CHAN_OFF
      SetNibble (out, CHAN_ON, &hwbuf);
      SetNibble (out, CHAN_OFF, &lwbuf);
    }
    else if (inp != 0)
    {
      //set corresponding nibble of loword to CHAN_ON, hiword to CHAN_OFF
      SetNibble (out, CHAN_ON, &lwbuf);
      SetNibble (out, CHAN_OFF, &hwbuf);
    }
    else
    {
      //set corresponding nibble of loword to CHAN_OFF, hiword to CHAN_OFF
      SetNibble (out, CHAN_OFF, &lwbuf);
      SetNibble (out, CHAN_OFF, &hwbuf);
    }

    dwbuf = MAKELONG (lwbuf, hwbuf);
    if ((ecode = LoadXP (index, dwbuf)))
    {
      return ecode;		//NB: we have not modified .xpdata
    }
  }

  //write the actual input#=inp-1
  if (inp > 8)
  {
    //set corresponding nibble of hiword to input#, loword already set to CHAN_OFF
    SetNibble (out, (BYTE) (inp - 9), &hwbuf);
  }
  else if (inp != 0)
  {
    //set corresponding nibble of loword to input#, hiword already set to CHAN_OFF
    SetNibble (out, (BYTE) (inp - 1), &lwbuf);
  }

  dwbuf = MAKELONG (lwbuf, hwbuf);

  if ((ecode = LoadXP (index, dwbuf)))
  {

    return ecode;		//NB: we have not modified .xpdata
  }

  //If we got here, everything is OK, so modify .xpdata
  board[index].xpdata = dwbuf;
  printk("leaving SetXP dwbuf = %lx\n", dwbuf);

  return 0;
}

static void SetNibble (BYTE nibblenumber, BYTE data, WORD * pWord)
{
  //Sets nibble #nibblenumber (0-3) (0=MSN, 3=LSN) in WORD pointed to by pWord
  //to value = data & 0xf
  WORD                      andmask, ormask, buffer;

  andmask = ~(0xf << (3 - nibblenumber) * 4);
  ormask = (data & 0xf) << (3 - nibblenumber) * 4;
  buffer = *pWord;
  buffer = ((buffer & andmask) | ormask);
  *pWord = buffer;
  return;
}

static void CreateRPS (bufinfo* bf,  MODE* pmode)
{
  //Create RPS programs for jpeg and bitmap capture for a given buffer

  DWORD                    *pRPS, label, *RPSstartlog, RPSstartphys;
  int                       outlines, outpixels, bytesperpix;
  DWORD                     bufferstart, protectionaddr, byteswap;
  DWORD                     linepix;
  int                       pitch;


  if (pmode->format == FORMAT_NTSC)
  {
    outlines =
      (pmode->bmpsize ==
       BMP_SIZE_LRG) ? BMPLINES_NTSC_LRG : BMPLINES_NTSC_SML;
    outpixels =
      (pmode->bmpsize ==
       BMP_SIZE_LRG) ? BMPPIXELS_NTSC_LRG : BMPPIXELS_NTSC_SML;
    linepix = SCALER_INP_LINES_PIX_N;
  }
  else
  {
    outlines =
      (pmode->bmpsize == BMP_SIZE_LRG) ? BMPLINES_PAL_LRG : BMPLINES_PAL_SML;
    outpixels =
      (pmode->bmpsize ==
       BMP_SIZE_LRG) ? BMPPIXELS_PAL_LRG : BMPPIXELS_PAL_SML;
    linepix = SCALER_INP_LINES_PIX_P;
  }

  byteswap = DMA_NOSWAP;
  if (pmode->bmpcolor == COLOR_YCRCB)
  {
    bytesperpix = BYTESPERPIX_YUV;
  }
  else if (pmode->bmpcolor == COLOR_YCRCB_BS)
  {
    bytesperpix = BYTESPERPIX_YUV;
    byteswap = DMA_SWAP2;
  }
  else if (pmode->bmpcolor == COLOR_RGB)
  {
    bytesperpix = BYTESPERPIX_RGB;
  }
  else
  {
    bytesperpix = BYTESPERPIX_MONO;
  }

  if (pmode->bmpflip)
  {
    bufferstart = (outlines - 1) * outpixels * bytesperpix;
    pitch = -outpixels * bytesperpix;
    protectionaddr = 0;
  }
  else
  {
    bufferstart = 0;
    pitch = outpixels * bytesperpix;
    protectionaddr = outlines * outpixels * bytesperpix;
  }

  //Bitmap (RPS task 0)
   pRPS = (DWORD*)(bf->rpsbmpuser);        

  *pRPS++ = RPS_CLRSIGNAL | RPS_O_FIDB | RPS_E_FIDB | RPS_VBIB;
  *pRPS++ = RPS_SETSIGNAL | RPS_SIG0;	//signals RPS0 active
  *pRPS++ = RPS_LDREG | (1 << 8) | (P_LCTS0 >> 2);
  *pRPS++ = outlines << 16;	//target line counter threshold
  *pRPS++ = RPS_LDREG | (6 << 8) | (P_BASEODD1 >> 2);	//BaseODD1,BaseEven1,ProtAddr1,Pitch1,PageMap,Format regs
  *pRPS++ = bufferstart;
  *pRPS++ = bufferstart;
  *pRPS++ = protectionaddr;
  *pRPS++ = pitch;
  *pRPS++ = (bf->ptbmpphys & 0xfffff000) | DMA_MAP | byteswap;
  *pRPS++ = linepix;
  *pRPS++ = RPS_UPLOAD | UPLD_ALLVIDEO;
  *pRPS++ = RPS_LDREG | (1 << 8) | (P_MC2 >> 2);
  *pRPS++ = UPLD_DMA1;
  *pRPS++ = RPS_PAUSE | RPS_OR | RPS_E_FIDB | RPS_O_FIDB;	//wait for even field start
  *pRPS++ = RPS_LDREG | (1 << 8) | (P_MC1 >> 2);	//enable transfer chan 1
  *pRPS++ = UPLD_TR_E_1;	//

  *pRPS++ = RPS_PAUSE | RPS_HT;	//wait until target lines event

  *pRPS++ = RPS_LDREG | (1 << 8) | (P_MC1 >> 2);	//stop transfer chan 1
  *pRPS++ = UPLD_TR_D_1;	//

  *pRPS++ = RPS_CLRSIGNAL | RPS_SIG0;	//signals RPS0 not active
  *pRPS++ = RPS_STOP;


  //Jpeg (RPS task 1)

  RPSstartphys = bf->rpsjpgphys;
  pRPS = (DWORD*)(bf->rpsjpguser); 



  RPSstartlog = pRPS;

  *pRPS++ = RPS_SETSIGNAL | RPS_SIG1;	//signals RPS1 active
  *pRPS++ = RPS_LDREG | (1 << 8) | (P_INTSTAT >> 2);
  *pRPS++ = ISR_DEBI_S | ISR_PIN3;

  //1st block
  *pRPS++ = RPS_LDREG | (4 << 8) | (P_DEBICFG >> 2);	//load reg's:
  *pRPS++ = DEBI_CFG_TIMEROFF | DEBI_CFG_16 | DEBI_CFG_FAST;	//CONFIG !!!DEBI_CFG_FAST
  *pRPS++ = DEBI_CMD_RD | (DEBI_BLOCK << 17) | A_ZORAN;	//COMMAND
  *pRPS++ = bf->ptjpgphys | DEBI_PAGE_ENABLE;	//PAGE
  *pRPS++ = 0;		//ADDRESS
  *pRPS++ = RPS_LDREG | (1 << 8) | (P_MC1 >> 2);	//enable..
  *pRPS++ = UPLD_TR_E_D;	//DEBI xfer ;
  *pRPS++ = RPS_CLRSIGNAL | RPS_DEBI;	//
  *pRPS++ = RPS_UPLOAD | RPS_DEBI_UPLD;	//upload DEBI ;
  *pRPS++ = RPS_PAUSE | RPS_DEBI;

  label = RPSstartphys + (DWORD) pRPS - (DWORD) RPSstartlog;	//phys addr of next instruction
  *pRPS++ = RPS_STREG | (1 << 8) | ((P_DEBIAD - 4) >> 2);	//store current DEBI addr
  *pRPS++ = (RPSstartphys + (DWORD) pRPS - (DWORD) RPSstartlog + 56);	//phys addr of ADDRESS <<<<<< (see below)

  //current block
  *pRPS++ = RPS_NOP;
  *pRPS++ = RPS_NOP;
  *pRPS++ = RPS_NOP;
  *pRPS++ = RPS_NOP;
  *pRPS++ = RPS_NOP;
  *pRPS++ = RPS_NOP;
  *pRPS++ = RPS_NOP;
  *pRPS++ = RPS_NOP;

  *pRPS++ = RPS_CLRSIGNAL | RPS_DEBI;
  *pRPS++ = RPS_LDREG | (4 << 8) | (P_DEBICFG >> 2);	//load reg's:
  *pRPS++ = DEBI_CFG_TIMEROFF | DEBI_CFG_16 | DEBI_CFG_FAST;	//CONFIG
  *pRPS++ = DEBI_CMD_RD | (DEBI_BLOCK << 17) | A_ZORAN;	//COMMAND
  *pRPS++ = bf->ptjpgphys | DEBI_PAGE_ENABLE;	//PAGE
  *pRPS++ = 0;		//ADDRESS     <<<<<<
  *pRPS++ = RPS_LDREG | (1 << 8) | (P_INTSTAT >> 2);
  *pRPS++ = ISR_DEBI_S;
  *pRPS++ = RPS_CLRSIGNAL | RPS_DEBI;
  *pRPS++ = RPS_UPLOAD | RPS_DEBI_UPLD;	//upload DEBI;
  *pRPS++ = RPS_PAUSE | RPS_DEBI;

  *pRPS++ = RPS_JUMP | RPS_GPIO3;	//
  *pRPS++ = label;

  *pRPS++ = RPS_LDREG | (1 << 8) | (P_MC1 >> 2);	//stop transfer DEBI
  *pRPS++ = UPLD_TR_D_D;	//
//    writemem( pRPS++, RPS_IRQ );

  *pRPS++ = RPS_CLRSIGNAL | RPS_SIG1;	//signals RPS1 not active
  *pRPS++ = RPS_STOP;

  return;
}

static int GetXP (cardinfo* boardinfo, int chan)
{
  //returns current xpoint channel # for a given channel #
  int                       shifts;
  int                       xpchan;

  chan &= 3;
  shifts = (3 - chan) * 4;

  if (((LOWORD (boardinfo->xpdata) >> shifts) & 0xf) == CHAN_OFF)
  {
    xpchan = 9 + ((HIWORD (boardinfo->xpdata) >> shifts) & 0x7);
  }
  else
  {
    xpchan = 1 + ((LOWORD (boardinfo->xpdata) >> shifts) & 0x7);
  }
  return xpchan;
}

static void StartRPS (int index, bufinfo* bf, int jpegon)
{
  WR7146 (index, P_RPSADDR0, bf->rpsbmpphys);
  WR7146 (index, P_RPSTO0, 0);
  WR7146 (index, P_RPSADDR1, bf->rpsjpgphys);
  WR7146 (index, P_RPSTO1, 0);
  if (jpegon)
  {
    WR7146 (index, P_MC1, EN_RPS1 | EN_DMA1_RPS0);	//
  }
  else
  {
    WR7146 (index, P_MC1, EN_DMA1_RPS0);
  }

  return;
}
static void CreateDIBHeader (bufinfo* bf, MODE * pMode)
{
  int                       i;
  DWORD                     width, height;

  if (pMode->format == FORMAT_NTSC)
  {
    width =
      (pMode->bmpsize ==
       BMP_SIZE_LRG) ? BMPPIXELS_NTSC_LRG : BMPPIXELS_NTSC_SML;
    height =
      (pMode->bmpsize ==
       BMP_SIZE_LRG) ? BMPLINES_NTSC_LRG : BMPLINES_NTSC_SML;
  }
  else
  {
    width =
      (pMode->bmpsize ==
       BMP_SIZE_LRG) ? BMPPIXELS_PAL_LRG : BMPPIXELS_PAL_SML;
    height =
      (pMode->bmpsize == BMP_SIZE_LRG) ? BMPLINES_PAL_LRG : BMPLINES_PAL_SML;
  }
  bf->lpbmi->bmiHeader.biSize =  sizeof (BITMAPINFOHEADER);
  bf->lpbmi->bmiHeader.biWidth = width;
  bf->lpbmi->bmiHeader.biHeight = height;
  bf->lpbmi->bmiHeader.biPlanes = 1;
  bf->lpbmi->bmiHeader.biCompression = 0;
  bf->lpbmi->bmiHeader.biSizeImage = 0;	//no need if no compression;
  bf->lpbmi->bmiHeader.biXPelsPerMeter =  0;
  bf->lpbmi->bmiHeader.biYPelsPerMeter =  0;
  if ((pMode->bmpcolor == COLOR_YCRCB) || (pMode->bmpcolor == COLOR_YCRCB_BS))
  {
    bf->lpbmi->bmiHeader.biBitCount = 16;
  }
  else
  {
    bf->lpbmi->bmiHeader.biBitCount = 24;
  }
  bf->lpbmi->bmiHeader.biClrUsed = 0;
  bf->lpbmi->bmiHeader.biClrImportant = 0;

  if (pMode->bmpcolor == COLOR_MONO)
  {
    //create grey scale color table
    for (i = 0; i < 256; i++)
    {
      bf->lpbmi->bmiColors[i].rgbBlue =	i;
      bf->lpbmi->bmiColors[i].rgbGreen = i;
      bf->lpbmi->bmiColors[i].rgbRed = i;
    }
    bf->lpbmi->bmiHeader.biBitCount = 8;
    bf->lpbmi->bmiHeader.biClrUsed = 256;
    bf->lpbmi->bmiHeader.biClrImportant = 0;
  }
  return;
}

static void WriteCaption (cardinfo* boardinfo, int chan, char *pCaption)
{
  int                       i;
  char                     *pCap = pCaption;
  int index = boardinfo->myminor;

  //Reset write addr counter
  WriteGA (index, A_CAPBUFARST, 0xff);

  //Write data
  for (i = 0; i < CAPTION_LENGTH; i++)
  {
    WriteGA (index, (WORD) (A_CAPBUF | CHAN[chan & 3] << 6), *pCap++);
  }
  return;
}

module_init (S615_init);
module_exit (S615_cleanup);
